from pathlib import Path
from sinter import Decoder, CompiledDecoder
from ilp_decoder.ilp import ILPDecoder
import numpy as np

import stim


class ILPSinterCompiledDecoder(CompiledDecoder):
    def __init__(self, decoder: ILPDecoder) -> None:
        self.decoder = decoder

    def decode_shots_bit_packed(
        self,
        *,
        bit_packed_detection_event_data: "np.ndarray",
    ) -> "np.ndarray":
        return self.decoder.decode_batch(
            shots=bit_packed_detection_event_data,
            bit_packed_shots=True,
            bit_packed_predictions=True,
        )


class ILPSinterDecoder(Decoder):
    def __init__(
        self,
        solver: str = "HIGHS",
        **solve_kwargs,
    ) -> None:
        """Class for decoding stim circuits with ILP solvers.

        Parameters
        ----------
        model : stim.DetectorErrorModel
            The detector error model of the stim circuit to be decoded
        solver : the solver to use for the ILP. The default is "HIGHS".
        solve_kwargs
            Additional keyword arguments are passed to `cvxpy.Problem.solve()`. See
            https://www.cvxpy.org/tutorial/solvers/index.html for more information.
        """
        self._solver = solver
        self._solve_kwargs = solve_kwargs

    def compile_decoder_for_dem(
        self, *, dem: stim.DetectorErrorModel
    ) -> CompiledDecoder:
        ilp = ILPDecoder(
            model=dem,
            solver=self._solver,
            **self._solve_kwargs,
        )
        return ILPSinterCompiledDecoder(ilp)

    def decode_via_files(
        self,
        *,
        num_shots: int,
        num_dets: int,
        num_obs: int,
        dem_path: Path,
        dets_b8_in_path: Path,
        obs_predictions_b8_out_path: Path,
        tmp_dir: Path,
    ) -> None:
        """Performs decoding by reading problems from, and writing solutions to, file paths.
        Args:
            num_shots: The number of times the circuit was sampled. The number of problems
                to be solved.
            num_dets: The number of detectors in the circuit. The number of detection event
                bits in each shot.
            num_obs: The number of observables in the circuit. The number of predicted bits
                in each shot.
            dem_path: The file path where the detector error model should be read from,
                e.g. using `stim.DetectorErrorModel.from_file`. The error mechanisms
                specified by the detector error model should be used to configure the
                decoder.
            dets_b8_in_path: The file path that detection event data should be read from.
                Note that the file may be a named pipe instead of a fixed size object.
                The detection events will be in b8 format (see
                https://github.com/quantumlib/Stim/blob/main/doc/result_formats.md ). The
                number of detection events per shot is available via the `num_dets`
                argument or via the detector error model at `dem_path`.
            obs_predictions_b8_out_path: The file path that decoder predictions must be
                written to. The predictions must be written in b8 format (see
                https://github.com/quantumlib/Stim/blob/main/doc/result_formats.md ). The
                number of observables per shot is available via the `num_obs` argument or
                via the detector error model at `dem_path`.
            tmp_dir: Any temporary files generated by the decoder during its operation MUST
                be put into this directory. The reason for this requirement is because
                sinter is allowed to kill the decoding process without warning, without
                giving it time to clean up any temporary objects. All cleanup should be done
                via sinter deleting this directory after killing the decoder.
        """
        dem = stim.DetectorErrorModel.from_file(dem_path)
        ilp = ILPDecoder(
            model=dem,
            solver=self._solver,
            **self._solve_kwargs,
        )
        shots = stim.read_shot_data_file(
            path=dets_b8_in_path,
            format="b8",
            num_detectors=dem.num_detectors,
            bit_packed=False,
        )
        predictions = ilp.decode_batch(shots)
        stim.write_shot_data_file(
            data=predictions,
            path=obs_predictions_b8_out_path,
            format="b8",
            num_observables=dem.num_observables,
        )


def sinter_decoders() -> dict[str, Decoder]:
    return {"ilp": ILPSinterDecoder()}
